<?php


namespace App\Services\refund;

use Api\PhpServices\Idgen\Idgen;
use Api\PhpUtils\Log\FileLog;
use App\Exception\BaseException;
use App\Models\Dictionary;
use App\Models\order\mysql\PayOrder;
use App\Models\order\mysql\PayOrderClearing;
use App\Models\order\mysql\PayOrderClearingItem;
use App\Models\order\mysql\PayOrderItem;
use App\Models\order\mysql\RefundOrder;
use App\Services\order\OrderService;

use App\Services\pingxx\PingxxService;
use Helpers\Strategy;

use App\Exception\custom\RefundException;
use Yaf\Application;

class RefundService
{

    /**
     * @param $order_item_id
     * @param $user_id
     * @throws RefundException
     * 发起退款请求
     */
    public function do_refund($order_item_id, $user_id) {

        if(empty($order_item_id)) {
            throw new RefundException(['cus' => 1]);
        }
        // 获取子单信息
        $ret = OrderService::getOrderItemData($order_item_id);
        if(empty($ret['result'])) {
            throw new RefundException(['cus' => 2]);
        }

        if(empty($ret['result']['user_id']) || $ret['result']['user_id'] != $user_id) {
            throw new RefundException(['cus' => 3]);
        }

        $order_item = $ret['result'];
        $where = [
            'order_item_id'=>$order_item_id,
            'user_id'=>$user_id,
            'order_id'=>$order_item['order_id'],
            'refund_amount'=>$order_item['payment'],
            'refund_order_id'=>$this->gen_refund_order_id($user_id),
        ];

        $data = $this->can_refund($where, $order_item);

        /*
        https://www.pingxx.com/api/%E8%AE%A2%E5%8D%95%20Refunds%20%E9%80%80%E6%AC%BE%E6%A6%82%E8%BF%B0.html
        订单退款 paid = true, status = canceled, amount_paid > 0 则表明可以退款
        amount_paid 大于 actual_amount 需要最后一单全退。
        具体见ping++ 文档
        */

        $pay = PingxxService::getInstance()->getOrder($data['pay_order']['third_order_id']);
        if(empty($pay['paid']) || $pay['paid'] != true) {
            throw new RefundException(['cus' => 12]);
        }

        // 存在多次支付的场景需要人工介入处理， 修复订单支付金额，退款金额
        if($pay['amount_paid'] > $pay['actual_amount']) {
            FileLog::error('pay-service: 订单可能存在多次成功支付，需人工确认处理。', json_encode($pay, JSON_UNESCAPED_UNICODE));
            throw new RefundException(['cus' => 13]);
        }


        // 构造退款数据
        $refund = [
            'id'=>$pay['id'], //pingxx订单ID
            'charge'=>$pay['charge'], //pingxx 支付订单ID
            'refund_mode' => 'to_source', //默认返回
            'charge_amount'=>$data['refund_order']['refund_amount'],
            'metadata' => [
                'app_id'=>'merchant-c',
                'environ'=>Application::app()->environ(),
                'refund_order_id'=>$data['refund_order']['refund_order_id'],
                'order_item_id'=>$data['refund_order']['order_item_id'],
                'goods_sku_id'=>$order_item['goods_sku_id'] ?? '',
            ],
        ];

        try{
            $ret = PingxxService::getInstance()->sendRefund($refund);
            if(!empty($ret["data"])) {
                $edit = ['request_pingxx_success_time'=>date('Y-m-d H:i:s')];
                RefundOrder::update($edit, ['refund_order_id'=>$data['refund_order']['refund_order_id']]);
            } else {
                throw new BaseException(['msg'=>$ret['error']['message'], 'code'=>'2011']);
            }

            return ['order_item_id'=>$order_item_id];

        }catch (\Exception $e) {
            $edit = ['refund_order_status'=>0];
            RefundOrder::update($edit, ['refund_order_id'=>$data['refund_order']['refund_order_id']]);
            throw $e;
        }
    }

    /**
     * @param $data
     * @return array
     * @throws RefundException
     * 退款回调
     */
    public function call_back($data) {

        if (!in_array($data['type'], ['refund.succeeded', 'order.refunded'])) {
            return ['error'=>'退款回调类型不正确'];
        }

        if (empty($data['data']['object'])) {
            return ['error'=>'退款回调格式不正确'];
        }

        $object = $data['data']['object'];
        if(empty($object['metadata']['app_id']) || $object['metadata']['app_id'] != "merchant-c") {
            //非支付中心发出的订单，需要忽略
            return ['error'=>'非merchant渠道订单，不做处理。'];
        }

        if($data['type'] == 'refund.succeeded') {
            $refund_order_id = $object['order_no'];
            $refunded = $object['succeed']; //"succeed": true,
        } else {
            $refund_order_id = $object['merchant_order_no'];
            $refunded = $object['refunded'];
        }

        if($refunded != true) {
            return ['error'=>'退款状态不正确，不做处理。'];
        }

        if(empty($object['metadata']['refund_order_id'])) {
            return ['error'=>'退款订单缺少matadata，不做处理。'];
        }

        // 修改状态，锁定业务
        $refund_order_id = $object['metadata']['refund_order_id'];
        $edit = [
            'refund_order_status'=>2, //pingxx回调成功,
            'pingxx_callback_success_time'=>date('Y-m-d H:i:s'), //回调成功时间
            'notify_account_status'=>1, //不需要通知钱包
        ];

        $cnt = RefundOrder::update($edit, ['refund_order_id'=>$refund_order_id, 'refund_order_status'=>1]);
        //保持幂等操作
        if($cnt == 0) {
            return ['refund_order_id'=>$refund_order_id];
        }

        return ['refund_order_id'=>$refund_order_id];

        //不用搞清分核算，是否需要平台记账
    }

    /**
     * @param $where
     * @return \Api\PhpUtils\Mysql\MysqlBase|array
     * @throws RefundException
     * 判断订单是否可退，判断是否存在退款申请
     */
    private function can_refund($where, $order_item) {

        /**
         * 业务方自行解决，订单关闭、订单超过时长不允许退款等逻辑
         * 支付中心退款判断仅做主单是否支付，主单是否支持退款的判断
         */

        try{
            RefundOrder::beginTransaction();
            $pay_order_item = PayOrderItem::getMaster('*', ['order_item_id'=>$where['order_item_id']]);
            if(empty($pay_order_item)) {
                throw new RefundException(['cus' => 4]);
            }
            if($pay_order_item['can_notify_account'] == 1) {
                throw new RefundException(['cus' => 5]);
            }

            $refund_order = RefundOrder::getMaster('*', ['order_item_id'=>$where['order_item_id']]);
            if(!empty($refund_order)) {
                if($refund_order['refund_order_status'] == Dictionary::REFUND_ORDER_STATUS_OK) {
                    throw new RefundException(['cus' => 6]);
                }

                if($refund_order['refund_order_status'] == Dictionary::REFUND_ORDER_STATUS_SEND) {
                    throw new RefundException(['cus' => 7]);
                }

            } else {
                $refund_order = $where;
                $refund_order['refund_order_status'] = Dictionary::REFUND_ORDER_STATUS_UNDO;
                //$refund_order['refund_amount'] = 0;
                $cnt = RefundOrder::insert($refund_order, ['rowCount'=>true]);
                if($cnt == 0) {
                    throw new RefundException(['cus' => 0]);
                }
            }

            //重新计算微信手续费
            $cleared_amount = $cleared_tip = 0;

            $pay_order = PayOrder::getMaster('*', ['order_id'=>$where['order_id']]);
            // 总共手续费
            $wx_clear = PayOrderClearing::getMaster('*', [
                'order_id'=>$where['order_id'],
                'pay_order_id'=>$pay_order['pay_order_id'],
                'pay_sub_type'=>406 //三方渠道手续费
                ]
            );

            // 计算已退金额、已退手续费
            $list = RefundOrder::selectMaster('*', ['order_id'=>$where['order_id']]);
            foreach ($list as $r) {
                if($r['refund_order_status'] == Dictionary::REFUND_ORDER_STATUS_UNDO) {
                    continue;
                }

                $cleared_amount += $r['refund_amount'];
                $cleared_tip += $r['refund_wx_tip'];
            }

            $wx_total_tip = 0;
            if($wx_clear) {
                $wx_total_tip = $wx_clear['pay_amount'];
            }

            $wx_tip = Strategy::getWxTip($order_item['payment'], $pay_order['pay_amount'], $cleared_amount, $wx_total_tip, $cleared_tip);

            $edit = [
                'refund_order_status'=>Dictionary::REFUND_ORDER_STATUS_SEND,
                'refund_amount'=>$order_item['payment'],
                'refund_wx_tip'=>$wx_tip,
            ];
            $ret = RefundOrder::update($edit, ['refund_order_id'=>$refund_order['refund_order_id']]);
            if(!$ret) {
                throw new RefundException(['cus' => 0]);
            }
            //重新获取
            $refund_order = RefundOrder::getMaster('*', ['refund_order_id'=>$refund_order['refund_order_id']]);
            $data = [
                'refund_order'=>$refund_order,
                'pay_order'=>$pay_order,
            ];

            RefundOrder::commit();

        }catch (\PDOException $e) {
            RefundOrder::rollback();
            throw new BaseException(['msg'=>$e->getMessage(), 'code'=>$e->getCode()]);
        }

        return $data;
    }

    private function gen_refund_order_id($user_id)
    {
        $number = substr($user_id, -2);
        $ids = $this->get_idgen_id($number, 1);
        return array_pop($ids);
    }

    private function get_idgen_id($number, $count = 1)
    {
        $number = intval($number);
        $res = Idgen::get(appConfig('idgen.partner'), appConfig('idgen.key'), [], [[
            "type" => "55c768",
            'number' => $number,
            "count" => $count]]);
        $id = $res['id_datetime']['55c768'] ?? [];
        return $id;
    }

}